/* Circuit's Cryptography Program v1.0
 6/13/2012 - 6/17/2012
 
 This program uses my "brick wall encryption" to make data unreadable.  The name "brick wall" refers to the brickwork-like pattern that it follows.  It makes 4
 passes over the file: once over the byte addresses, once between the byte addresses, then again over the byte addresses, then again between.  To make this more
 confusing, the first and last passes rotate bytes character-wise, and the middle passes rotate bytes bitwise.  This produces ciphertext that is utterly
 unreadable to humans, and it's complicated enough that I don't think a human or machine could analyze it for patterns (although I can't say it's impossible).
 However, since the password "wraps" through the file in a predictable way, an enemy could write a very efficient brute force attack program to crack a file
 (although this would require complete knowledge of the algorithm, and the determination to write such a program).  As with any encryption program, the best
 strategy is to encrypt a file several times using several different lengthy passwords, and don't leave a note saying "this file encrypted with Circuit's
 Cryptography Program".  Another good trick is to encrypt using decrypt and then decrypt using encrypt.
 
 Need to add:
 "Pyramid" algorithm
 An algorithm which derives its series of encryption & decryption method calls from the password
 Transposition (this is super important!)
 Rotating alphabets
 
 Dev notes:
 strlen() isn't used to get the length of the bytes array because it triggers a warning re: signage.  I think it's actually safe, but I won't risk it.
 Need to check if the file is blank.
 */

#include <stdio.h>
#include <stdlib.h>																		/*needed for exit() and malloc()*/
#include <string.h>																		/*needed for strchr()*/
#include <unistd.h>																		/*needed for getcwd()*/

#define NUM_BYTES_TO_PRINT 100

void getString (char *, int);
FILE *getFile (char *fileName);
long int getSize (FILE *);
unsigned char *readFile (FILE *, long int, unsigned char *);
void saveFile (unsigned char *, long int, char *);										/*function prototypes*/
void printBytes (unsigned char *, long int, long int);
unsigned char rotateByte (unsigned char, int);
int bitwiseVigenereEncrypt (unsigned char *, long int, long int, char *, int);
int bitwiseVigenereDecrypt (unsigned char *, long int, long int, char *, int);
int bitwiseVigenereEncryptBB (unsigned char *, long int, long int, char *, int);
int bitwiseVigenereDecryptBB (unsigned char *, long int, long int, char *, int);
int stepBack (char *, int);
int stepForward (char *, int);															/*Why is this sea blue?!  I hope this isn't a reserved word...*/
void rotateBetweenBytes(unsigned char *, int, int);										/*nvm, it's not sea blue anymore, and I'm not using it anyway*/
unsigned char rotateChar (unsigned char, int);
int VigenereEncrypt (unsigned char *, long int, long int, char *, int);
int VigenereDecrypt (unsigned char *, long int, long int, char *, int);
void rotateBetweenChars(unsigned char *, int, int);
int VigenereEncryptBB(unsigned char *, long int, long int, char *, int);
int VigenereDecryptBB(unsigned char *, long int, long int, char *, int);
void BrickWallEncrypt(unsigned char *, long int, long int, char *, int);				/*Multi-pass encryption algorithm which calls the 4 Vigenere functions.*/
void BrickWallDecrypt(unsigned char *, long int, long int, char *, int);				/*The matching decryption algorithm.*/


int main (int argc, const char * argv[]) {
	
	/*File-related Declarations*/
	char fileName[255];																	/*space for the filename*/
	char password[255];																	/*space for the password*/
	FILE *inFile = NULL;																/*file pointer*/
	long int fileSize;																	/*file size (in bytes)*/
	unsigned char *bytes;																/*pointer for copy of file in memory*/
	int passwordStep = 0;																/*the index of the next letter in the sequence*/
	char mode[32];																		/*mode of operation (encrypt or decrypt)*/
	char dir[1024];																		/*space for dir name (this could be malloc'd...)*/
	
	printf("\nCircuit's Cryptography Program v1.0\n");
	printf("current working directory is %s\n", getcwd(dir, sizeof dir));
	
	while (mode[0] != 'd' && mode[0] != 'e')											/*Require an 'e' or 'd' as input.*/
	{
		printf("What would you like to do? ('e' for encrypt, 'd' for decrypt): ");		/*It signifies the mode of operation.*/
		getString(mode, sizeof mode);													/*Get a 1-character string (i.e. a char).*/
	}
	
	while (inFile == NULL)																/*Require a valid/accessible file.*/
	{
		printf("Please input the name of the file to ");
		
		switch (mode[0])
		{
			case 'e':
				printf("encrypt: ");
				break;
				
			case 'd':
				printf("decrypt: ");
				break;
				
			default:
				printf("\n... oh no, I don't know what I'm supposed to do!!!\n");
				printf("How did I get here?!  MOMMY!!!\n");
				exit(1);																/*This won't happen.*/
				break;
		}

		getString(fileName, sizeof fileName);											/*Get filename.*/
		inFile = getFile(fileName);														/*Attempt to open file.*/
		if (inFile == NULL)
		{
			printf("Sorry, '%s' couldn't be opened.  ", fileName);
			printf("It may be open in another program, or you may have mistyped its name.\n");
		}
	}/*if the file couldn't be opened, the loop will continue*/
	
	
	/*Count the file's length, allocate space in memory, and read it in.*/
	fileSize = getSize (inFile);														/*Get file size (in bytes).*/
	bytes = readFile(inFile, fileSize, bytes);											/*Read file in and copy its bytes beginning at &bytes.*/
	fclose(inFile);																		/*Close the file (it isn't needed anymore).*/
	printf("%s opened successfully.\n", fileName);
	
	printf("Please input the password: ");
	getString(password, sizeof password);												/*Get password.*/
	if (strlen(password) == 0)															/*If the user didn't enter a password (just hit Enter/Return),*/
	{
		password[0] = 'H';																/*use a default password.  Hebrews 4:13*/
		password[1] = 'e';
		password[2] = 'b';
		password[3] = '4';
		password[4] = ':';
		password[5] = '1';
		password[6] = '3';
		password[7] = '\0';
		printf("Default password was used: %s\n\n", password);
	}
	
	printf("File contents (first %d bytes):\n", NUM_BYTES_TO_PRINT);
	printBytes(bytes, fileSize, NUM_BYTES_TO_PRINT);

	
	/*Perform the encryption.*/
	switch (mode[0])
	{
		case 'e':
			BrickWallEncrypt(bytes, 0, fileSize, password, passwordStep);
			printf("\nEncrypted:\n");
			break;
			
		case 'd':
			BrickWallDecrypt(bytes, 0, fileSize, password, passwordStep);
			printf("\nDecrypted:\n");
			break;
	}
	printBytes(bytes, fileSize, NUM_BYTES_TO_PRINT);
	
	
	/*Save the file.*/
	printf("\nSave the file as: ");
	getString(fileName, sizeof fileName);
	saveFile(bytes, fileSize, fileName);												/*Save the encrypted/decrypted copy of the file.*/

	free(bytes);																		/*Release the memory that was allocated for it.*/
	
	printf("Saved as '%s' in %s", fileName, dir);										/*Tell the user where it is.*/
	
    return 0;																			/*We're done!*/
}/*main*/

void getString (char *stringBuffer, int bufferSize) {
	char *newLine;																		/*pointer to newline character*/
	
	fgets(stringBuffer, bufferSize, stdin);												/*accept string input*/
																						/*sizeof doesn't work here (I guess getString doesn't know it)*/
	newLine = strchr(stringBuffer, '\n');												/*search string buffer for newline character*/
	if (newLine != NULL)																/*if newline was found*/
	{
		*newLine = '\0';																/*replace newline with null (strings are null-terminated)*/
	}
}/*getString*/

FILE *getFile (char *fileName) {
	FILE *myFile;																		/*file pointer*/
	
	myFile = fopen(fileName, "r");														/*attempt to open the file*/
	
	return myFile;																		/*return the file pointer (this will be NULL on failure)*/
}/*getFile*/

long int getSize (FILE *myFile) {

	long int length;
	
	fseek(myFile, 0L, SEEK_END);														/*go to the end of the file*/
	length = ftell(myFile);																/*record the position relative to the beginning (length of file)*/
	
	rewind(myFile);																		/*go back to the start of the file*/
	
	return length;																		/*return the length of the file*/
}/*getSize*/

void printBytes (unsigned char *bytes, long int fileSize, long int charsToPrint) {

	if (fileSize > charsToPrint)														/*don't print the entire file, unless it's a short one*/
	{
		for (int i = 0; i < charsToPrint; i++)
		{
			printf("%c", bytes[i]);
		}
		printf("...\n");
	}
	else
	{
		for (int i = 0; i <= fileSize; i++)
		{
			printf("%c", bytes[i]);
		}
		printf("\n");
	}
}/*printBytes*/

unsigned char rotateByte (unsigned char byte, int places) {
	unsigned char rotated = byte;														/*just in case it doesn't change*/
	
	places = places % 8;																/*take the remainder to remove any 8's*/
	
	if (places < 0)																		/*if the rotation is negative (left)*/
	{
		places *= -1;																	/*the places value must be positive when used for either direction*/
		rotated = (byte << places) | (byte >> (8 - places) );
	}
	else																				/*if the rotation is positive (right)*/
	{
		if (places > 0)
		{
			rotated = (byte >> places) | (byte << (8 - places) );
		}/*endif*/
	}/*endif*/
	
	return rotated;
}/*rotateByte*/

unsigned char *readFile (FILE *inFile, long int fileSize, unsigned char *bytes) {
	
	bytes = (unsigned char *) malloc(fileSize * sizeof(unsigned char));					/*allocate memory to hold a copy of the file*/
	
	if (bytes == (unsigned char *) NULL)
	{
		printf("Not enough memory to read in the file.\n");
		exit(1);
	}
	
	fread(bytes, 1, fileSize, inFile);													/*(array pointer, num bytes per unit, num units, file pointer)*/
	
	return bytes;
}/*readFile*/

void saveFile (unsigned char *bytes, long int fileSize, char *fileName) {
	
	FILE *outFile;
	
	outFile = fopen(fileName, "w");														/*create output file*/
	
	fwrite(bytes, 1, fileSize, outFile);												/*write bytes to file (var ptr, #bytes/unit, #units, file ptr)*/
	
	fclose(outFile);																	/*close output file*/
}/*saveFile*/

int bitwiseVigenereEncrypt(unsigned char *bytes, long int start, long int end, char *password, int passwordStep) {
	
	/*Warning: this method currently doesn't check to make sure that the start and end are inside the bounds of bytes[].
	Also note: the end value was originally the file size.  The true end value is ONE LESS than the given end value!*/
	
	int passwordLength = strlen(password);
	
	for (int i = start; i < end; i++)
	{
		bytes[i] = rotateByte(bytes[i], password[passwordStep]);
		
		passwordStep++;
			
		if (passwordStep == passwordLength)												/*passwordLength is 1 greater than the last index of password[]*/
		{
			passwordStep = 0;
		}
		
	}
	return passwordStep;																/*wanted pwS to be a global, but that didn't work out*/
}/*bitwiseVigenereEncrypt*/

int bitwiseVigenereDecrypt(unsigned char *bytes, long int start, long int end, char *password, int passwordStep) {
	
	/*Same notes about the encryption method apply to the decrypt.
	This method moves backwards from the end to the start.*/
	
	int passwordLength = strlen(password);
	

	for (int i = end-1; i >= start; i--)
	{
		bytes[i] = rotateByte(bytes[i], -password[passwordStep]);
		passwordStep--;
		
		if (passwordStep == -1)
		{
			passwordStep = passwordLength - 1;
		}
	}
	
	return passwordStep;
}/*bitwiseVigenereDecrypt*/

int stepBack (char *password, int passwordStep) {
	
	if (passwordStep == 0)																/*if the step is at the first letter*/
	{
		passwordStep = strlen(password) - 1;											/*move it to the last letter*/
	}
	else {
		passwordStep--;																	/*if not, just subtract 1*/
	}
	
	return passwordStep;
}/*stepBack*/

int stepForward (char *password, int passwordStep) {
	
	/*This function needs to be tested.*/
	
	if (passwordStep == strlen(password) - 1)											/*if the step is at the last letter*/
	{
		passwordStep = 0;																/*move it to the first letter*/
	}
	else {
		passwordStep++;																	/*if not, just add 1*/
	}
	
	return passwordStep;
}/*stepForward*/

void rotateBetweenBytes(unsigned char *bytes, int index, int places) {
	
	/*Warning: this function can potentially go outside of the bounds of the *bytes array.  Use with care.*/
	
	unsigned char left, right, fused;
	int indexPlus1 = index + 1;
	
	left = bytes[index];
	right = bytes[indexPlus1];															/*this is the dangerous part!*/
	
	left = left << 4;																	/*rotate left; least significant become most significant*/
	right = right >> 4;																	/*vice versa*/
	
	fused = left + right;																/*right half of left and left half of right*/
	
	fused = rotateByte(fused, places);													/*rotate the "fused byte"*/
	
	bytes[index] = bytes[index] >> 4;
	bytes[index] = bytes[index] << 4;
	bytes[index] += fused >> 4;
	
	bytes[indexPlus1] = bytes[indexPlus1] << 4;
	bytes[indexPlus1] = bytes[indexPlus1] >> 4;
	bytes[indexPlus1] += fused << 4;
	
}/*rotateBetweenBytes*/

int bitwiseVigenereEncryptBB(unsigned char *bytes, long int start, long int end, char *password, int passwordStep) {
	
	/*Warning: this method currently doesn't check to make sure that the start and end are inside the bounds of bytes[].
	Double warning: this method goes BETWEEN bytes, so you need to be extra careful.
	Also note: the end parameter was originally the file size.  The true end value is ONE LESS than the given end value!
	Wait, actually it's 2 less.
	For a full-file encrypt, the index of the last byte should be the length of the file -2.
	(Subt. 1 to convert the file length (natural number) to the last array index (0-based), then subt. 1 more to hit the 2nd-to-last index.)*/
	
	int passwordLength = strlen(password);
	
	for (int i = start; i < (end - 1); i++)
	{
		rotateBetweenBytes(bytes, i, password[passwordStep]);
		passwordStep++;
		if (passwordStep == passwordLength)
		{
			passwordStep = 0;
		}
	}
	return passwordStep;
}/*bitwiseVigenereEncryptBB*/

int bitwiseVigenereDecryptBB(unsigned char *bytes, long int start, long int end, char *password, int passwordStep) {
	
	int passwordLength = strlen(password);
	
	for (int i = end-2; i >= start; i--)												/*count backwards from the end*/
	{
		rotateBetweenBytes(bytes, i, -password[passwordStep]);
		passwordStep--;
		
		if (passwordStep == -1)
		{
			passwordStep = passwordLength - 1;
		}
	}
	
	return passwordStep;
}/*bitwiseVigenereDecryptBB*/

unsigned char rotateChar (unsigned char byte, int places) {
	
	/*This method rotates a byte "letterwise", through the ASCII char set, by adding or subtracting a value (places).
	It's short because it depends on the machine to do things, specifically, to "wrap" the high and low ends of a char,
	so that when a char goes over 255, it goes to 0 and continues counting forwards.  And likewise, when a char goes below 0,
	it goes to 255 and continues counting backwards from there.*/
	
/*	places = places % 256;																this is unnecessary and slows down the program*/
	
	byte += places;
	
	return byte;
}/*rotateChar*/

int VigenereEncrypt(unsigned char *bytes, long int start, long int end, char *password, int passwordStep) {
	
	/*Warning: this method currently doesn't check to make sure that the start and end are inside the bounds of bytes[].
	Also note: the end value is the file size.  Like array subscripts, the subscript of the last value is ONE LESS than the file size!
	That's why the for loop goes until i < end.*/
	
	int passwordLength = strlen(password);
	
	for (int i = start; i < end; i++)
	{
		bytes[i] = rotateChar(bytes[i], password[passwordStep]);
		
		passwordStep++;
		
		if (passwordStep == passwordLength)
		{
			passwordStep = 0;
		}
	}
	return passwordStep;
}/*VigenereEncrypt*/

int VigenereDecrypt(unsigned char *bytes, long int start, long int end, char *password, int passwordStep) {
	
	int passwordLength = strlen(password);
	
	for (int i = end-1; i >= start; i--)
	{
		bytes[i] = rotateChar(bytes[i], -password[passwordStep]);
		passwordStep--;
		
		if (passwordStep == -1)
		{
			passwordStep = passwordLength - 1;
		}
	}
	
	return passwordStep;
}/*VigenereDecrypt*/

void rotateBetweenChars(unsigned char *bytes, int index, int places) {
	
	/*Warning: this function can potentially go outside of the bounds of the *bytes array.  Use with care.*/
	
	unsigned char left, right, fused;
	
	int indexPlus1 = index + 1;
	
	left = bytes[index];
	right = bytes[indexPlus1];
	
	left = left << 4;
	right = right >> 4;
	
	fused = left + right;
	
	fused = rotateChar(fused, places);
	
	bytes[index] = bytes[index] >> 4;
	bytes[index] = bytes[index] << 4;
	bytes[index] += fused >> 4;
	
	bytes[indexPlus1] = bytes[indexPlus1] << 4;
	bytes[indexPlus1] = bytes[indexPlus1] >> 4;
	bytes[indexPlus1] += fused << 4;
	
}/*rotateBetweenChars*/

int VigenereEncryptBB(unsigned char *bytes, long int start, long int end, char *password, int passwordStep) {
	
	/*This method doesn't check bounds.  It goes Between Bytes.  Yadda yadda yadda.  See bitwiseVigenereEncryptBB() for more details.*/
	
	int passwordLength = strlen(password);
	
	for (int i = start; i < (end - 1); i++)
	{
		rotateBetweenChars(bytes, i, password[passwordStep]);
		passwordStep++;
		if (passwordStep == passwordLength)
		{
			passwordStep = 0;
		}
	}
	return passwordStep;
}/*VigenereEncryptBB*/

int VigenereDecryptBB(unsigned char *bytes, long int start, long int end, char *password, int passwordStep) {
	
	int passwordLength = strlen(password);
	
	for (int i = end-2; i >= start; i--)
	{
		rotateBetweenChars(bytes, i, -password[passwordStep]);
		passwordStep--;
		
		if (passwordStep == -1)
		{
			passwordStep = passwordLength - 1;
		}
	}
	
	return passwordStep;
}/*VigenereDecryptBB*/

void BrickWallEncrypt(unsigned char *bytes, long int start, long int end, char *password, int passwordStep) {
	
	passwordStep = VigenereEncrypt(bytes, 0, end, password, passwordStep);
	printf("\nFirst pass complete...\n");
	passwordStep = bitwiseVigenereEncryptBB(bytes, 0, end, password, passwordStep);
	printf("Second pass complete...\n");
	passwordStep = bitwiseVigenereEncrypt(bytes, 0, end, password, passwordStep);
	printf("Third pass complete...\n");
	passwordStep = VigenereEncryptBB(bytes, 0, end, password, passwordStep);
	printf("All passes completed.\n");
	
}/*BrickWallEncrypt*/

void BrickWallDecrypt(unsigned char *bytes, long int start, long int end, char *password, int passwordStep) {
	
	/*Formula for determining the ending password step in a piece of BrickWall'd text.
	The number of steps is (number of chars * 2) + ((number of chars - 1) * 2)
	Because the regular Vigenere methods step through the entire length of the file, and the bitwise ones step through the length - 1.*/
	
	/*This kind of math would also be used for a brute-force attack to predict where a guess password "wraps" in a file.  An enemy could use this
	 knowledge to attack a part of a file instead of the whole file, and test more passwords in less time.  This is a serious weakness that makes
	 cracking highly feasible for files which are only encrypted once.  Once transposition is implemented, this won't be as much of a problem.*/
	
	/*If ever I re-use this code to only encode partial files, I will need to re-write this code.
	The way this formula is coded assumes that the whole file will be decrypted!*/
	passwordStep = ((end*2) + ((end-1)*2)) % strlen(password);
	passwordStep = stepBack(password, passwordStep);
	
	passwordStep = VigenereDecryptBB(bytes, 0, end, password, passwordStep);
	printf("\nFirst pass complete...\n");
	passwordStep = bitwiseVigenereDecrypt(bytes, 0, end, password, passwordStep);
	printf("Second pass complete...\n");
	passwordStep = bitwiseVigenereDecryptBB(bytes, 0, end, password, passwordStep);
	printf("Third pass complete...\n");
	passwordStep = VigenereDecrypt(bytes, 0, end, password, passwordStep);
	printf("All passes completed.\n");
	
}/*BrickWallDecrypt*/
